
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/proc.h>
#include <sys/timeout.h>
#include <sys/timebase.h>

#include <machine/bus.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcidevs.h>

#include "elanvar.h"
#include "elanreg.h"
#include "elan_mmcr.h"

#ifdef COUNTER_TIMEBASE
struct elan_timer_softc;
struct elan_gp_timebase {
	struct timebase		egp_timebase;
	struct elan_timer_softc	*egp_sc;
	volatile u_int64_t	egp_count;
};
#endif /* COUNTER_TIMEBASE */

struct elan_timer_softc {
	struct device		sc_dev;
	bus_space_tag_t		et_mmcrt;
	bus_space_handle_t	et_mmcr;
#ifdef COUNTER_TIMEBASE
	struct elan_gp_timebase	et_tb;
#endif /* COUNTER_TIMEBASE */
};

//static void elan_timer_cal(struct elan_timer_softc *);
static int elan_timer_probe(struct device *, void *, void *);
static void elan_timer_attach(struct device *, struct device *, void *);
#ifdef COUNTER_TIMEBASE
static u_int64_t elan_gp_tb_read(struct timebase *);
static u_int64_t elan_gp_tb_update(struct timebase *);
#endif /* COUNTER_TIMEBASE */

struct cfattach elan_timer_ca = {
        sizeof(struct elan_timer_softc), elan_timer_probe, elan_timer_attach
};

struct cfdriver elan_timer_cd = {
        0, "elan_timer", DV_DULL
};

static int
elan_timer_probe(struct device *parent, void *match, void *aux)
{
	return (1);
}

static void
elan_timer_attach(struct device *parent, struct device *self, void *aux)
{
	struct elan_timer_softc *sc = (struct elan_timer_softc *)self;
	struct elan_attach_args *ea = aux;
	u_int8_t tmp;
	int s = splimp();

	sc->et_mmcrt = ea->elan_memt;
	sc->et_mmcr = ea->elan_mem;

	tmp = bus_space_read_1(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_SWTMRCFG);

	bus_space_write_1(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_SWTMRCFG, 1);

	printf((tmp & 1) ? ": 33.000MHz clock" : ": 33.333MHz clock");	

	/* Disable GP2, clear the current count, and set period to max */
	bus_space_write_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR2CTL,
		ELAN_GPTMRCTL_P_ENB_WR | ELAN_GPTMRCTL_CONT_CMP);
	bus_space_write_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR2CNT, 0);
	bus_space_write_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR2MAXCMPA, 0);

	/* Disable GP1, clear the current count, and set period to max */
	bus_space_write_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR1CTL,
		ELAN_GPTMRCTL_P_ENB_WR |
		ELAN_GPTMRCTL_CONT_CMP |
		ELAN_GPTMRCTL_PSC_SEL |
		ELAN_GPTMRCTL_RTG);
	bus_space_write_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR1CNT, 0);
	bus_space_write_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR1MAXCMPA, 0);

	/* Enable GP2 */
	bus_space_write_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR2CTL,
		ELAN_GPTMRCTL_ENB | ELAN_GPTMRCTL_P_ENB_WR | ELAN_GPTMRCTL_CONT_CMP);

	/* Enable GP1 */
	bus_space_write_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR1CTL,
		ELAN_GPTMRCTL_ENB |
		ELAN_GPTMRCTL_P_ENB_WR |
		ELAN_GPTMRCTL_CONT_CMP |
		ELAN_GPTMRCTL_PSC_SEL |
		ELAN_GPTMRCTL_RTG);

#ifdef COUNTER_TIMEBASE
	sc->et_tb.egp_sc = sc;
	sc->et_tb.egp_timebase.tb_read = elan_gp_tb_read;
	sc->et_tb.egp_timebase.tb_update = elan_gp_tb_update;
#ifdef SOEKRIS_X1_FREQ
	sc->et_tb.egp_timebase.tb_frequency = (SOEKRIS_X1_FREQ + 2) / 4;
#else
	sc->et_tb.egp_timebase.tb_frequency = (tmp & 1) ?
	    (33000000 / 4) : (33333333 / 4);
#endif /* SOEKRIS_X1_FREQ */
	timebase_register(&sc->et_tb.egp_timebase);

	if(0) {
		int i;
		for(i = 0; i < 100; ++i) {
			DELAY(i * 11000);
			printf("timer = 0x%16.16llx\n",
				sc->et_tb.egp_timebase.tb_update(&sc->et_tb.egp_timebase));
		}
	}
			
#endif /* COUNTER_TIMEBASE */
	//elan_timer_cal(sc);

	splx(s);
}

#if 0
static __inline void _rtcout(int reg, u_int8_t data);

static __inline void
_rtcout(int reg, u_int8_t data)
{
	outb(0x70, reg);
	outb(0x71, data);
}

static __inline u_int8_t _rtcin(int reg);

static
__inline u_int8_t _rtcin(int reg)
{
	outb(0x70, reg);
	return inb(0x71);
}

static __inline u_int8_t get_sec(void);

static __inline u_int8_t
get_sec() 
{
	while(_rtcin(0x0a) & 0x80)
		;
 	return _rtcin(0x00);	
}

static void
elan_timer_cal(struct elan_timer_softc *sc)
{
	u_int16_t milli, micro;
	u_int16_t best_milli = 0xffff, best_micro = 0xffff;
	u_int8_t sec1;
	u_int8_t sec2 = get_sec();
	int i;

	do {
		sec1 = get_sec();
	} while(sec1 == sec2);

	/* Clear millisecond count */
	bus_space_read_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_SWTMRMILLI);

	for(i = 0; i < 5; ++i) {
		do {
			sec2 = get_sec();
		} while(sec1 == sec2);

		milli = bus_space_read_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_SWTMRMILLI);
		micro = bus_space_read_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_SWTMRMICRO);

		sec1 = sec2;

		if(milli < best_milli || (milli == best_milli && micro < best_micro)) {
			best_milli = milli;
			best_micro = micro;
		}
	}

	printf(" 1s = %dmilli %dmicro", best_milli, best_micro);
}
#endif  /* 0 */

#ifdef COUNTER_TIMEBASE

static u_int64_t
elan_gp_tb_read(struct timebase *tb)
{
	struct elan_gp_timebase *etb = (struct elan_gp_timebase *)tb;
	struct elan_timer_softc *sc = etb->egp_sc;
	u_int64_t count1, count2;
	u_int32_t gp1a, gp1b, gp2;
	u_int32_t cnt, tmp;

	do {	
		gp1a = bus_space_read_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR1CNT);
		gp2 = bus_space_read_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR2CNT);
		gp1b = bus_space_read_2(sc->et_mmcrt, sc->et_mmcr, ELAN_MMCR_GPTMR1CNT);
	} while(gp1a != gp1b);

	do {
		count1 = etb->egp_count;
		count2 = etb->egp_count;
	} while(count1 != count2);

	cnt = (gp1a << 16) | gp2;

	tmp = (u_int32_t)count1;
	count2 = (count1 & 0xffffffff00000000ull) | cnt;
	if(tmp > cnt)
		count2 += 1ull << 32;

	return (count2);
}

static u_int64_t
elan_gp_tb_update(struct timebase *tb)
{
	struct elan_gp_timebase *etb = (struct elan_gp_timebase *)tb;
	u_int64_t count;
	int s;

	count = elan_gp_tb_read(tb);

	s = splhigh();
	etb->egp_count = count;
	splx(s);

	return (count);
}
#endif /* COUNTER_TIMEBASE */

